import threading, pickle
from . import db, index, cache
from .db import txnfun

__all__ = ["environment", "datastore", "autostore"]

class environment(object):
    def __init__(self, *, path=None, getpath=None, recover=False):
        if path is not None:
            self.path = path
            self.getpath = None
        else:
            self.path = None
            self.getpath = getpath
        self.recover = recover
        self.lk = threading.Lock()
        self.bk = None

    def __call__(self):
        with self.lk:
            if self.bk is None:
                if self.path is None:
                    self.path = self.getpath()
                self.bk = db.environment(self.path, recover=self.recover)
            return self.bk

    def close(self):
        with self.lk:
            if self.bk is not None:
                self.bk.close()
                self.bk = None

class storedesc(object):
    pass

def storedescs(obj):
    t = type(obj)
    ret = getattr(t, "__didex_attr", None)
    if ret is None:
        ret = []
        for nm, val in t.__dict__.items():
            if isinstance(val, storedesc):
                ret.append((nm, val))
        t.__didex_attr = ret
    return ret

class datastore(object):
    def __init__(self, name, *, env=None, path=".", ncache=None):
        self.name = name
        self.lk = threading.Lock()
        if env:
            self.env = env
        else:
            self.env = environment(path=path)
        self._db = None
        if ncache is None:
            ncache = cache.cache()
        self.cache = ncache
        self.cache.load = self._load

    def db(self):
        with self.lk:
            if self._db is None:
                self._db = self.env().db(self.name)
            return self._db

    def _load(self, id):
        try:
            return pickle.loads(self.db().get(id))
        except:
            raise KeyError(id, "could not unpickle data")

    def _encode(self, obj):
        return pickle.dumps(obj)

    def get(self, id, *, load=True):
        return self.cache.get(id, load=load)

    @txnfun(lambda self: self.db().env.env)
    def register(self, obj, *, tx):
        id = self.db().add(self._encode(obj), tx=tx)
        for nm, attr in storedescs(obj):
            attr.register(id, obj, tx)
        self.cache.put(id, obj)
        return id

    @txnfun(lambda self: self.db().env.env)
    def unregister(self, id, *, vfy=None, tx):
        obj = self.get(id)
        if vfy is not None and obj is not vfy:
            raise RuntimeError("object identity crisis: " + str(vfy) + " is not cached object " + obj)
        for nm, attr in storedescs(obj):
            attr.unregister(id, obj, tx)
        self.db().remove(id, tx=tx)
        self.cache.remove(id)

    @txnfun(lambda self: self.db().env.env)
    def update(self, id, *, vfy=None, tx):
        obj = self.get(id, load=False)
        if vfy is not None and obj is not vfy:
            raise RuntimeError("object identity crisis: " + str(vfy) + " is not cached object " + obj)
        for nm, attr, in storedescs(obj):
            attr.update(id, obj, tx)
        self.db().replace(id, self._encode(obj), tx=tx)

class autotype(type):
    def __call__(self, *args, **kwargs):
        new = super().__call__(*args, **kwargs)
        new.id = self.store.register(new)
        return new

class autostore(object, metaclass=autotype):
    def __init__(self):
        self.id = None

    def save(self):
        self.store.update(self.id, vfy=self)

    def remove(self):
        self.store.unregister(self.id, vfy=self)
        self.id = None
